# -*- coding: utf-8 -*-

from datetime import timezone
import re
import csv
import emoji
from lbsnstructure.lbsnstructure_pb2 import *
from google.protobuf.timestamp_pb2 import Timestamp
import datetime
import logging
import time
from collections import Counter
from json import JSONDecoder, JSONDecodeError
# for debugging only:
from google.protobuf import text_format
from shapely import geos, wkb, wkt
# https://gis.stackexchange.com/questions/225196/conversion-of-a-geojson-into-ewkb-format
geos.WKBWriter.defaults['include_srid'] = True

class HelperFunctions():

    @staticmethod
    def get_version():
        """Gets the program version number from version file in root"""
        with open('VERSION') as version_file:
            version_var = version_file.read().strip()
        return version_var

    @staticmethod
    def log_main_debug(debug_text):
        """Issues a main debug log in case it is needed for static functions."""
        logging.getLogger('__main__').debug(debug_text)

    @staticmethod
    def null_notice(x):
        """Reporting: Suppresses null notice (for Null island) if value is zero."""
        null_notice_x = lambda x: f'(Null Island: {x})' if x > 0 else ''
        return null_notice_x(x)

    @staticmethod
    def utc_to_local(utc_dt):
        return utc_dt.replace(tzinfo=timezone.utc).astimezone(tz=None)

    @staticmethod
    def cleanhtml(raw_html):
        cleanr = re.compile('<.*?>')
        cleantext = re.sub(cleanr, '', raw_html)
        return cleantext

    @staticmethod
    def extract_emoji(str):
        #str = str.decode('utf-32').encode('utf-32', 'surrogatepass')
        #return list(c for c in str if c in emoji.UNICODE_EMOJI)
        return list(c for c in str if c in emoji.UNICODE_EMOJI)

    @staticmethod
    def getRectangleBounds(points):
        lats = []
        lngs = []
        for point in points:
            lngs.append(point[0])
            lats.append(point[1])
        limYMin = min(lats)
        limYMax = max(lats)
        limXMin = min(lngs)
        limXMax = max(lngs)
        return limYMin,limYMax,limXMin,limXMax

    @staticmethod
    def createNewLBSNRecord_with_id(record,id,origin):
            # initializes new record with composite ID
            c_Key = CompositeKey()
            c_Key.origin.CopyFrom(origin)
            c_Key.id = id
            record.pkey.CopyFrom(c_Key)
            return record

    @staticmethod
    def createNewLBSNRelationship_with_id(lbsnRelationship,relation_to_id, relation_from_id, relation_origin):
            # initializes new relationship with 2 composite IDs for one origin
            c_Key_to = CompositeKey()
            c_Key_to.origin.CopyFrom(relation_origin)
            c_Key_to.id = relation_to_id
            c_Key_from = CompositeKey()
            c_Key_from.origin.CopyFrom(relation_origin)
            c_Key_from.id = relation_from_id
            r_Key = RelationshipKey()
            r_Key.relation_to.CopyFrom(c_Key_to)
            r_Key.relation_from.CopyFrom(c_Key_from)
            lbsnRelationship.pkey.CopyFrom(r_Key)
            return lbsnRelationship

    @staticmethod
    def isPostReaction(jsonString):
        if 'quoted_status' in jsonString or 'retweeted_status' in jsonString or jsonString.get('in_reply_to_status_id_str'):
            # The retweeted field will return true if a tweet _got_ retweeted
            # To detect if a tweet is a retweet of other tweet, check the retweeted_status field
            return True
        else:
            return False

    @staticmethod
    def assignMediaPostType(jsonMediaString):
        # if post, get type of first entity
        typeString = jsonMediaString[0].get('type')
        # type is either photo, video, or animated_gif
        # https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/extended-entities-object.html
        if typeString:
            if typeString == "photo":
                post_type = lbsnPost.IMAGE
            elif typeString in ("video","animated_gif"):
                post_type = lbsnPost.VIDEO

        else:
            post_type = lbsnPost.OTHER
            log.debug(f'Other Post type detected: {jsonMediaString}')
        return post_type

    @staticmethod
    def parseJSONDateStringToProtoBuf(jsonDateString):
        # Parse String -Timestamp Format found in Twitter json
        dateTimeRecord = datetime.datetime.strptime(jsonDateString,'%a %b %d %H:%M:%S +0000 %Y')
        protobufTimestampRecord = Timestamp()
        # Convert to ProtoBuf Timestamp Recommendation
        protobufTimestampRecord.FromDatetime(dateTimeRecord)
        return protobufTimestampRecord

    @staticmethod
    def parse_csv_datestring_to_protobuf(csv_datestring):
        # Parse String -Timestamp Format found in Flickr csv
        dateTimeRecord = datetime.datetime.strptime(csv_datestring,'%m/%d/%Y %H:%M:%S')
        protobufTimestampRecord = Timestamp()
        # Convert to ProtoBuf Timestamp Recommendation
        protobufTimestampRecord.FromDatetime(dateTimeRecord)
        return protobufTimestampRecord

    @staticmethod
    def getMentionedUsers(userMentions_jsonString,origin):
        mentionedUsersList = []
        for userMention in userMentions_jsonString:    #iterate over the list
            refUserRecord = HelperFunctions.createNewLBSNRecord_with_id(lbsnUser(),userMention.get('id_str'),origin)
            refUserRecord.user_fullname = userMention.get('name') # Needs to be saved
            refUserRecord.user_name = userMention.get('screen_name')
            mentionedUsersList.append(refUserRecord)
        return mentionedUsersList

    @staticmethod
    def substituteReferencedUser(mainPost, origin, log):
        # Look for mentioned userRecords
        refUser_pkey = None
        userMentionsJson = mainPost.get('entities').get('user_mentions')
        if userMentionsJson:
            refUserRecords = HelperFunctions.getMentionedUsers(userMentionsJson,origin)
            # if it is a retweet, and the status contains 'RT @', and the mentioned UserID is in status, we can almost be certain that it is the userid who posted the original tweet that was retweeted
            if refUserRecords and refUserRecords[0].user_name.lower() in mainPost.get('text').lower() and  mainPost.get('text').startswith(f'RT @'):
                refUser_pkey = refUserRecords[0].pkey
            if refUser_pkey is None:
                log.warning(f'No User record found for referenced post in: {mainPost}')
                input("Press Enter to continue... (post will be saved without userid)")
        return refUser_pkey

    @staticmethod
    def null_check(recordAttr):
        """Helper function to check for Null Values
        """
        if not recordAttr:
            return None
        else:
            # This function will also remove Null bytes from string, which aren't supported by Postgres
            if isinstance(recordAttr, str):
                recordAttr = HelperFunctions.clean_null_bytes_from_str(recordAttr)
            return recordAttr

    @staticmethod
    def null_check_datetime(recordAttr):
        if not recordAttr:
            return None
        else:
            try:
                dt_attr = recordAttr.ToDatetime()
            except:
                return None
            if dt_attr == datetime.datetime(1970, 1, 1, 0, 0, 0):
                return None
            else:
                return recordAttr.ToDatetime()

    ## EWKB Conversion now in-code, not server-side
    #def geoconvertOrNone(geom):
    #    if geom:
    #        return "extensions.ST_GeomFromText(%s,4326)"
    #    else:
    #        return "%s"

    @staticmethod
    def returnEWKBFromGeoTEXT(text):
        if not text:
            return None
        geom = wkt.loads(text)
        geos.lgeos.GEOSSetSRID(geom._geom, 4326)
        geom = geom.wkb_hex
        return geom

    @staticmethod
    def decode_stacked(document, pos=0, decoder=JSONDecoder()):
        NOT_WHITESPACE = re.compile(r'[^\s]')
        while True:
            match = NOT_WHITESPACE.search(document, pos)
            if not match:
                return
            pos = match.start()

            try:
                obj, pos = decoder.raw_decode(document, pos)
            except JSONDecodeError:
                # do something sensible if there's some error
                raise
            yield obj

    def clean_null_bytes_from_str(str):
        str_without_null_byte = str.replace('\x00','')
        return str_without_null_byte

    @staticmethod
    def merge_existing_records(oldrecord, newrecord):
        # Basic Compare function for GUIDS
        # First check if length of both ProtoBuf Messages are the same
        oldRecordString = oldrecord.SerializeToString()
        newRecordString = newrecord.SerializeToString()
        if not len(oldRecordString) == len(newRecordString):
            # no need to do anything if same lengt
            oldrecord.MergeFrom(newrecord)
            #updatedrecord = self.deepCompareMergeMessages(oldrecord,newrecord)

    @staticmethod
    def load_importer_mapping_module(origin):
        """ Switch import module based on origin input
            1 - Instagram, 2 - Flickr, 3 - Twitter
        """
        if origin == 2:
            from .field_mapping_flickr import FieldMappingFlickr as importer
        elif origin == 3:
            from .field_mapping_twitter import FieldMappingTwitter as importer
        return importer

    @staticmethod
    def dict_type_switcher(desc_name):
        """ Create protoBuf messages by name
        """
        dict_switcher = {
            lbsnCountry().DESCRIPTOR.name: lbsnCountry(),
            lbsnCity().DESCRIPTOR.name: lbsnCity(),
            lbsnPlace().DESCRIPTOR.name: lbsnPlace(),
            lbsnUser().DESCRIPTOR.name: lbsnUser(),
            lbsnUserGroup().DESCRIPTOR.name:  lbsnUserGroup(),
            lbsnPost().DESCRIPTOR.name: lbsnPost(),
            lbsnPostReaction().DESCRIPTOR.name: lbsnPostReaction(),
            lbsnRelationship().DESCRIPTOR.name: lbsnRelationship()
        }
        return dict_switcher.get(desc_name)

    @staticmethod
    def check_notice_empty_post_guid(post_guid):
        if not post_guid:
           self.log.warning(f'No PostGuid\n\n{jsonStringDict}')
           input("Press Enter to continue... (entry will be skipped)")
           return False
        else:
           return True

class LBSNRecordDicts():
    def __init__(self):
        self.lbsnCountryDict = dict()
        self.lbsnCityDict = dict()
        self.lbsnPlaceDict = dict()
        self.lbsnUserGroupDict = dict()
        self.lbsnUserDict = dict()
        self.lbsnPostDict = dict()
        self.lbsnPostReactionDict = dict()
        self.lbsnRelationshipDict = dict()
        self.KeyHashes = {lbsnPost.DESCRIPTOR.name: set(),
                         lbsnCountry.DESCRIPTOR.name: set(),
                         lbsnCity.DESCRIPTOR.name: set(),
                         lbsnPlace.DESCRIPTOR.name: set(),
                         lbsnUserGroup.DESCRIPTOR.name: set(),
                         lbsnUser.DESCRIPTOR.name: set(),
                         lbsnPostReaction.DESCRIPTOR.name: set(),
                         lbsnRelationship.DESCRIPTOR.name: set()}
        self.CountGlob = 0
        # returns all recordsDicts in correct order, with names as references (tuple)
        self.allDicts = [
            (self.lbsnCountryDict,lbsnCountry().DESCRIPTOR.name),
            (self.lbsnCityDict,lbsnCity().DESCRIPTOR.name),
            (self.lbsnPlaceDict,lbsnPlace().DESCRIPTOR.name),
            (self.lbsnUserGroupDict,lbsnUserGroup().DESCRIPTOR.name),
            (self.lbsnUserDict,lbsnUser().DESCRIPTOR.name),
            (self.lbsnPostDict,lbsnPost().DESCRIPTOR.name),
            (self.lbsnPostReactionDict,lbsnPostReaction().DESCRIPTOR.name),
            (self.lbsnRelationshipDict,lbsnRelationship().DESCRIPTOR.name)
            ]

    def getTypeCounts(self):
        countList = []
        for x, y in self.KeyHashes.items():
            countList.append(f'{x}: {len(y)} ')
        return ''.join(countList)

    def update_keyHash(self, record):
        # Keep lists of pkeys for each type
        # this can be used to check for duplicates or to get a total count for each type of records (Number of unique Users, Countries, Places etc.)
        # in this case we assume that origin_id remains the same in each program iteration!
        if record.DESCRIPTOR.name == lbsnRelationship().DESCRIPTOR.name:
            # we need the complete uuid of both entities for relationships because they can span different origin_ids
            self.KeyHashes[record.DESCRIPTOR.name].add(f'{record.pkey.relation_to.origin.origin_id}{record.pkey.relation_to.id}{record.pkey.relation_from.origin.origin_id}{record.pkey.relation_from.id}{record.relationship_type}')
        else:
            # all other entities can be globally uniquely identified by their local guid
            self.KeyHashes[record.DESCRIPTOR.name].add(record.pkey.id)

    def deepCompareMergeMessages(self,oldRecord,newRecord):
        # Do a deep compare
        # ProtoBuf MergeFrom does a fine job
        # only problem is it concatenates repeate strings, which may result in duplicate entries
        # we take care of this prior to submission (see submitData classes)
        oldRecord.MergeFrom(newRecord)
        #for descriptor in oldRecord.DESCRIPTOR.fields:
        #        if descriptor.label == descriptor.LABEL_REPEATED:
        #            if value_old == value_new:
        #                return oldRecord
        #            elif not value_old:
        #                newEntries = value_new
        #            else:
        #                # only add difference (e.g. = new values)
        #                newEntries = list(set(value_new) - set(value_old))
        #            x = getattr(oldRecord, descriptor.name)
        #            x.extend(newEntries)
        return oldRecord

    def AddRecordsToDict(self,records):
        if isinstance(records,(list,)):
            for record in records:
                self.AddRecordToDict(record)
        else:
            record = records
            self.AddRecordToDict(record)

    def dictSelector(self, record):
        """ Get dictionary by type name
        """
        dictSwitcher = {
            lbsnPost().DESCRIPTOR.name: self.lbsnPostDict,
            lbsnCountry().DESCRIPTOR.name: self.lbsnCountryDict,
            lbsnCity().DESCRIPTOR.name: self.lbsnCityDict,
            lbsnPlace().DESCRIPTOR.name: self.lbsnPlaceDict,
            lbsnPostReaction().DESCRIPTOR.name: self.lbsnPostReactionDict,
            lbsnUser().DESCRIPTOR.name: self.lbsnUserDict,
            lbsnUserGroup().DESCRIPTOR.name: self.lbsnUserGroupDict
        }
        return dictSwitcher.get(record.DESCRIPTOR.name)

    def AddRecordToDict(self,newrecord):
        dict = self.dictSelector(newrecord)
        pkeyID = newrecord.pkey.id
        if newrecord.pkey.id in dict:
            oldrecord = dict[pkeyID]
            # oldrecord will be modified/updated
            HelperFunctions.merge_existing_records(oldrecord,newrecord)
        else:
            # just count new entries
            self.countProgressReport()
            self.update_keyHash(newrecord) # update keyHash only necessary for new record
            dict[pkeyID] = newrecord

    def countProgressReport(self):
        self.CountGlob += 1
        if self.CountGlob % 1000 == 0:
            # progress report (modulo)
            print(f'Identified Output Records {self.CountGlob}..                                                    ', end='\r')
            sys.stdout.flush()

    def AddRelationshipToDict(self,newrelationship):
        pkeyID = f'{newrelationship.pkey.relation_to.origin.origin_id}{newrelationship.pkey.relation_to.id}{newrelationship.pkey.relation_from.origin.origin_id}{newrelationship.pkey.relation_from.id}{newrelationship.relationship_type}'
        if not pkeyID in self.lbsnRelationshipDict:
            self.countProgressReport()
            self.lbsnRelationshipDict[pkeyID] = newrelationship
            self.update_keyHash(newrelationship) # update keyHash only necessary for new record

class GeocodeLocations():
    def __init__(self):
        self.geocodeDict = dict()

    def load_geocodelist(self,file):
        with open(file, newline='', encoding='utf8') as f: #read each unsorted file and sort lines based on datetime (as string)
            #next(f) #Skip Headerrow
            locationfile_list = csv.reader(f, delimiter=',', quotechar='', quoting=csv.QUOTE_NONE)
            for location_geocode in locationfile_list:
                self.geocodeDict[location_geocode[2].replace(';',',')] = (float(location_geocode[0]),location_geocode[1]) # lat/lng
        print(f'Loaded {len(self.geocodeDict)} geocodes.')

class TimeMonitor():
    def __init__(self):
        self.now = time.time()

    def stop_time(self):
        later = time.time()
        hours, rem = divmod(later-self.now, 3600)
        minutes, seconds = divmod(rem, 60)
        difference = int(later - self.now)
        reportMsg = f'{int(hours):0>2} Hours {int(minutes):0>2} Minutes and {seconds:05.2f} Seconds passed.'
        return reportMsg

class MemoryLeakDetec():
    # use this class to identify memory leaks
    # execute .before() and .after() to see the difference in new objects being added
    # execute report to list
    # if there are high numbers of specific types of objects, use printType to print these, e.g. .printType(list)
    # see also http://tech.labs.oliverwyman.com/blog/2008/11/14/tracing-python-memory-leaks/
    os = __import__('os')
    def __init__(self):
        global defaultdict
        global get_objects
        from collections import defaultdict
        from gc import get_objects
        self._before = defaultdict(int)
        self._after = defaultdict(int)

    def before(self):
        for i in get_objects():
            self._before[type(i)] += 1

    def after(self):
        for i in get_objects():
            self._after[type(i)] += 1

    def report(self):
        reportStat = ""
        if self._before and self._after:
            reportStat = [(k, self._after[k] -  self._before[k]) for k in self._after if self._after[k] -  self._before[k]]
        return reportStat

    def printType(self, type, max = 100):
        x = 0
        toplist = []
        for obj in get_objects():
            if isinstance(obj, type):
                x += 1
                if x < max:
                    toplist.append(obj)
        print(f'Count all of Type {type}: {x}. Top {max}: {toplist}')


# -*- coding: utf-8 -*-
import argparse
import os
import sys

class BaseConfig():
    def __init__(self):
        ## Set Default Config options here
        ## or define options as input args
        self.Origin = 3 # Defaults to 3: Twitter (1 - Instagram, 2 - Flickr, 3 - Twitter)
        self.is_local_input = False # Read from File/CSV
        self.local_file_type = 'json' # If localread, specify filetype (json, csv etc.)
        self.InputPath = None # optionally provide path to input folder, otherwise ./Input/ will be used
        self.is_stacked_json = False
        self.dbUser_Input = 'example-user-name'
        self.dbPassword_Input = 'example-user-password'
        self.dbServeradressInput = '222.22.222.22'
        self.dbNameInput = 'test_db2'
        self.dbUser_Output = None #'example-user-name2'
        self.dbPassword_Output = None #'example-user-password2'
        self.dbServeradressOutput = None #'111.11.11.11'
        self.dbNameOutput = None #'test_db'
        self.transferlimit = None
        self.transferCount = 50000 #default:50k # after how many parsed records should the result be transferred to the DB. Larger values improve speed, because duplicate check happens in Python and not in Postgres Coalesce; larger values are heavier on memory.
        self.number_of_records_to_fetch = 10000
        self.transferReactions = True
        # Enable this option in args to prevent empty posts stored due to Foreign Key Exists Requirement
        self.disableReactionPostReferencing = False # 0 = Save Original Tweets of Retweets in "posts"; 1 = do not store Original Tweets of Retweets; !Not implemented: 2 = Store Original Tweets of Retweets as "post_reactions"
        self.ignore_non_geotagged = False
        self.startWithdb_row_number = 0
        self.end_with_db_row_number = None
        self.debugMode = 'INFO' #needs to be implemented
        self.geocodeLocations = False # provide path to CSV file with location geocodes (CSV Structure: lat, lng, name)
        self.ignore_input_source_list = False # Provide a list of input_source types that will be ignored (e.g. to ignore certain bots etc.)
        self.input_lbsn_type = None # Input type, e.g. "post", "profile", "friendslist", "followerslist" etc.
        self.MapRelations = False # Set to true to map full relations, e.g. many-to-many relationships such as user_follows, user_friend, user_mentions etc. are mapped in a separate table
        self.CSVOutput = False # Set to True to Output all Submit values to CSV
        self.CSVsuppressLinebreaks = True # Set to False will not remove intext-linebreaks (\r or \n) in output CSVs
        self.recursiveLoad = False
        self.skip_until_file = "" # If local input, skip all files until file with name x appears (default: start immediately)

    def parseArgs(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('-sO', "--Origin", default=self.Origin)
        parser.add_argument('-lI', "--LocalInput", action='store_true', default=False)
        parser.add_argument('-lT', "--LocalFileType", default=self.local_file_type)
        parser.add_argument('-iP', "--InputPath", default=self.InputPath)
        parser.add_argument('-iS', "--isStackedJson", action='store_true', default=False)
        parser.add_argument('-pO', "--dbPassword_Output", default=self.dbPassword_Output)
        parser.add_argument('-uO', "--dbUser_Output", default=self.dbUser_Output)
        parser.add_argument('-aO', "--dbServeradressOutput", default=self.dbServeradressOutput)
        parser.add_argument('-nO', "--dbNameOutput", default=self.dbNameOutput)
        parser.add_argument('-pI', "--dbPassword_Input", default=self.dbPassword_Input)
        parser.add_argument('-uI', "--dbUser_Input", default=self.dbUser_Input)
        parser.add_argument('-aI', "--dbServeradressInput", default=self.dbServeradressInput)
        parser.add_argument('-nI', "--dbNameInput", default=self.dbNameInput)
        parser.add_argument('-t', "--transferlimit", default=self.transferlimit)
        parser.add_argument('-tC', "--transferCount", default=self.transferCount)
        parser.add_argument('-nR', "--numberOfRecordsToFetch", default=self.number_of_records_to_fetch)
        parser.add_argument('-tR', "--disableTransferReactions", action='store_true')
        parser.add_argument('-rR', "--disableReactionPostReferencing", action='store_true', default=False)
        parser.add_argument('-iG', "--ignoreNonGeotagged", action='store_true', default=self.ignore_non_geotagged)
        parser.add_argument('-rS', "--startWithDBRowNumber", default=self.startWithdb_row_number)
        parser.add_argument('-rE', "--endWithDBRowNumber", default=self.end_with_db_row_number)
        parser.add_argument('-d', "--debugMode", default=self.debugMode)
        parser.add_argument('-gL', "--geocodeLocations", default=self.geocodeLocations)
        parser.add_argument('-igS', "--ignoreInputSourceList", default=self.ignore_input_source_list)
        parser.add_argument('-iT', "--inputType", default=self.input_lbsn_type)
        parser.add_argument('-mR', "--mapFullRelations", default=self.MapRelations)
        parser.add_argument('-CSV', "--CSVOutput", action='store_true', default=self.CSVOutput)
        parser.add_argument('-CSVal', "--CSVallowLinebreaks", action='store_true', default=False)
        parser.add_argument('-rL', "--recursiveLoad", action='store_true', default=False)
        parser.add_argument('-sF', "--skipUntilFile", default=self.skip_until_file)

        args = parser.parse_args()
        if args.LocalInput:
            self.is_local_input = True
            self.local_file_type = args.LocalFileType
            if args.isStackedJson:
                self.is_stacked_json = True
            if not args.InputPath:
                self.InputPath = f'{os.getcwd()}\\01_Input\\'
                print(f'Using Path: {self.InputPath}')
            else:
                if args.InputPath.endswith("\\"):
                    input_path = args.InputPath
                else:
                    input_path = f'{args.InputPath}\\'
                self.InputPath = input_path
        else:
            self.dbUser_Input = args.dbUser_Input
            self.dbPassword_Input = args.dbPassword_Input
            self.dbServeradressInput = args.dbServeradressInput
            self.dbNameInput = args.dbNameInput
        if args.Origin:
            self.Origin = int(args.Origin)
        if args.geocodeLocations:
            self.geocodeLocations = f'{os.getcwd()}\\{args.geocodeLocations}'
        if args.ignoreInputSourceList:
            self.ignore_input_source_list = f'{os.getcwd()}\\{args.ignoreInputSourceList}'
        if args.dbUser_Output:
            self.dbUser_Output = args.dbUser_Output
            self.dbPassword_Output = args.dbPassword_Output
            self.dbServeradressOutput = args.dbServeradressOutput
            self.dbNameOutput = args.dbNameOutput
        if args.transferlimit:
            self.transferlimit = int(args.transferlimit)
            if self.transferlimit == 0:
                self.transferlimit = None
        if args.transferCount:
            self.transferCount = int(args.transferCount)
        if args.numberOfRecordsToFetch:
            self.number_of_records_to_fetch = int(args.numberOfRecordsToFetch)
        if args.disableTransferReactions is True:
            self.transferReactions = False
        if args.disableReactionPostReferencing:
            self.disableReactionPostReferencing = True
        self.ignore_non_geotagged = args.ignoreNonGeotagged
        if args.startWithDBRowNumber:
            self.startWithdb_row_number = int(args.startWithDBRowNumber)
        if args.endWithDBRowNumber:
            self.end_with_db_row_number = int(args.endWithDBRowNumber)
        self.debugMode = args.debugMode
        if args.inputType:
            self.input_lbsn_type = args.inputType
        if args.mapFullRelations:
            self.MapRelations = True
        if args.CSVOutput:
            self.CSVOutput = True
        if args.CSVallowLinebreaks:
            self.CSVsuppressLinebreaks = False
        if args.recursiveLoad:
            self.recursiveLoad = True
        if args.skipUntilFile:
            self.skip_until_file = args.skipUntilFile



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
lbsntransform package script to load, format and store data
from and to common lbsn structure

Import options:
    - Postgres Database or
    - local CSV/json/stacked json import
Output options:
    - Postgres Database or
    - local ProtoBuf and CSV Import (prepared for Postgres /Copy)
"""

__author__ = "Alexander Dunkel"
__license__ = "GNU GPLv3"

import logging
import io

def main():
    """ Main function to process data from postgres db or local file input
        to postgres db or local file output
    """
    import sys
    from .classes.helper_functions import TimeMonitor
    from .classes.helper_functions import HelperFunctions as HF
    from .classes.submit_data import LBSNTransfer
    from .classes.load_data import LoadData
    from .config.config import BaseConfig

    # Set Output to Replace in case of encoding issues (console/windows)
    sys.stdout = io.TextIOWrapper(sys.stdout.detach(), sys.stdout.encoding, 'replace')
    # Load Config, will be overwritten if args are given
    config = BaseConfig()
    # Parse args
    config.parseArgs()
    sys.stdout.flush()
    log = set_logger()
    # load import mapper depending on lbsn origin (e.g. 1 = Instagram, 2 = Flickr, 3 = Twitter)
    importer = HF.load_importer_mapping_module(config.Origin)
    # establish output connection
    conn_output, cursor_output = LoadData.initialize_output_connection(config)
    output = LBSNTransfer(dbCursor=cursor_output,
                          dbConnection=conn_output,
                          storeCSV=config.CSVOutput,
                          SUPPRESS_LINEBREAKS=config.CSVsuppressLinebreaks)
    # load from local json/csv or from PostgresDB
    if config.is_local_input:
        loc_filelist = LoadData.read_local_files(config)
    else:
        # establish input connection
        cursor_input = LoadData.initialize_input_connection(config)

    # start settings
    processed_records = 0
    processed_total = 0
    start_number = 0
    # Start Value, Modify to continue from last processing
    continue_number = config.startWithdb_row_number

    # Optional Geocoding
    geocode_dict = False
    if config.geocodeLocations:
        geocode_dict = LoadData.load_geocodes(config.geocodeLocations)
    # Optional ignore input sources
    if config.ignore_input_source_list:
        ignore_sources_set = LoadData.load_ignore_sources(config.ignore_input_source_list)

    finished = False
    # initialize field mapping structure
    import_mapper = importer(config.disableReactionPostReferencing,
                             geocode_dict,
                             config.MapRelations,
                             config.transferReactions,
                             config.ignore_non_geotagged,
                             ignore_sources_set)
    # Manually add entries that need submission prior to parsing data
    # add_bundestag_group_example(import_mapper)

    how_long = TimeMonitor()
    # loop input DB until transferlimit reached or no more rows are returned
    while not finished:
        if config.transferlimit:
            max_records = config.transferlimit - processed_total
        else:
            max_records = None
        if config.is_local_input:
            if continue_number > len(loc_filelist) - 1:
                break
            records = LoadData.fetch_data_from_file(loc_filelist,
                                                         continue_number,
                                                         config.is_stacked_json,
                                                         config.local_file_type)
            # skip empty files
            if not records:
                continue_number += 1
                continue
        else:
            records = LoadData.fetch_json_data_from_lbsn(cursor_input,
                                                         continue_number,
                                                         max_records,
                                                         config.number_of_records_to_fetch)
            if not records:
                break
        if config.is_local_input:
            continue_number += 1
        else:
            continue_number = records[-1][0] #last returned db_row_number
        processed_count, finished = LoadData.loop_input_records(records,
                                                                max_records,
                                                                import_mapper,
                                                                config)
        processed_records += processed_count
        processed_total += processed_count
        print(f'{processed_total} input records processed (up to {continue_number}). '
              f'Count per type: {import_mapper.lbsnRecords.getTypeCounts()}records.', end='\n')
        # update console
        # On the first loop or after 500.000 processed records, transfer results to DB
        if not start_number or processed_records >= config.transferCount or finished:
            sys.stdout.flush()
            print(f'Storing {import_mapper.lbsnRecords.CountGlob} records .. {HF.null_notice(import_mapper.null_island)})')
            output.storeLbsnRecordDicts(import_mapper)
            output.commitChanges()
            processed_records = 0
            ## create a new empty dict of records
            import_mapper = importer(config.disableReactionPostReferencing,
                                       geocode_dict,
                                       config.MapRelations)
        # remember the first processed DBRow ID
        if not start_number:
            if config.is_local_input:
                start_number = 1
            else:
                start_number = records[0][0] #first returned db_row_number

    # submit remaining
    # ??
    if import_mapper.lbsnRecords.CountGlob > 0:
        print(f'Transferring remaining {import_mapper.lbsnRecords.CountGlob} to db.. {HF.null_notice(import_mapper.null_island)})')
        output.storeLbsnRecordDicts(import_mapper)
        output.commitChanges()

    # finalize all transactions (csv merge etc.)
    output.finalize()

    # Close connections to DBs
    if not config.is_local_input:
        cursor_input.close()
    if config.dbUser_Output:
        cursor_output.close()
    log.info(f'\n\nProcessed {processed_total} input records '
             f'(Input {start_number} to {continue_number}).')
    input(f'Done. {how_long.stop_time()}')

def set_logger():
    """ Set logging handler manually, so we can also print to console while logging to file"""

    logging.basicConfig(handlers=[logging.FileHandler('log.log', 'w', 'utf-8')],
                        format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
                        datefmt='%H:%M:%S',
                        level=logging.DEBUG)
    log = logging.getLogger(__name__)
    # Get Stream handler
    logging.getLogger().addHandler(logging.StreamHandler())
    return log

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from .helper_functions import HelperFunctions
from .helper_functions import LBSNRecordDicts
from lbsnstructure.lbsnstructure_pb2 import *
from google.protobuf.timestamp_pb2 import Timestamp
import shapely.geometry as geometry
from shapely.geometry.polygon import Polygon
import logging
import re
# for debugging only:
from google.protobuf import text_format

class FieldMappingTwitter():
    def __init__(self,
                 disableReactionPostReferencing=False,
                 geocodes=False,
                 mapFullRelations=False,
                 map_reactions=True,
                 ignore_non_geotagged=False,
                 ignore_sources_set=set()):
        # We're dealing with Twitter in this class, lets create the OriginID globally
        # this OriginID is required for all CompositeKeys
        origin = lbsnOrigin()
        origin.origin_id = lbsnOrigin.TWITTER
        self.origin = origin
        self.lbsnRecords = LBSNRecordDicts() #this is where all the data will be stored
        self.null_island = 0
        self.log = logging.getLogger('__main__')#logging.getLogger()
        self.disableReactionPostReferencing = disableReactionPostReferencing
        self.mapFullRelations = mapFullRelations
        self.geocodes = geocodes
        self.map_reactions = map_reactions
        self.ignore_non_geotagged = ignore_non_geotagged
        self.ignore_sources_set = ignore_sources_set

    def parseJsonRecord(self, jsonStringDict, input_lbsn_type = None):
        # decide if main object is post or user json
        if input_lbsn_type and input_lbsn_type in ('friendslist', 'followerslist'):
            for user, relatedUserList in jsonStringDict.items():
                userRecord = HelperFunctions.createNewLBSNRecord_with_id(lbsnUser(),str(user),self.origin)
                self.lbsnRecords.AddRecordsToDict(userRecord)
                for relatedUser in relatedUserList:
                    relatedRecord = HelperFunctions.createNewLBSNRecord_with_id(lbsnUser(),str(relatedUser),self.origin)
                    self.lbsnRecords.AddRecordsToDict(relatedRecord)
                    # note the switch of order here, direction is important for 'isConnected', and the different list each give us a different view on this relationship
                    if input_lbsn_type == 'friendslist':
                        relationshipRecord = HelperFunctions.createNewLBSNRelationship_with_id(lbsnRelationship(), userRecord.pkey.id, relatedRecord.pkey.id, self.origin)
                    elif input_lbsn_type == 'followerslist':
                        relationshipRecord = HelperFunctions.createNewLBSNRelationship_with_id(lbsnRelationship(), relatedRecord.pkey.id, userRecord.pkey.id, self.origin)
                    relationshipRecord.relationship_type = lbsnRelationship.isCONNECTED
                    self.lbsnRecords.AddRelationshipToDict(relationshipRecord)
        elif (input_lbsn_type and input_lbsn_type == 'profile') or 'screen_name' in jsonStringDict:
            # user
            userRecord = self.extractUser(jsonStringDict)
            self.lbsnRecords.AddRecordsToDict(userRecord)
            #sys.exit(f'Post record: {text_format.MessageToString(userRecord,as_utf8=True)}')
            if not userRecord.is_private:
                # if user profile is private, we cannot access posts
                userStatus = None
                if 'status' in jsonStringDict:
                    userStatus = jsonStringDict.get('status')
                elif 'quoted_status' in jsonStringDict:
                    userStatus = jsonStringDict.get('quoted_status')
                elif 'retweeted_status' in jsonStringDict:
                    userStatus = jsonStringDict.get('retweeted_status')
                # in case user status is available
                if userStatus:
                    self.parseJsonPost(userStatus, userPkey = userRecord.pkey)
        else:
            self.parseJsonPost(jsonStringDict)

    def parseJsonPost(self, jsonStringDict, userPkey = None):
        # Post
        # 1. Extract all relevant Post Attributes
        #    1.a extract post coordinates
        #    1.b extract user attributes
        #    1.c extract place attributes (poi, city, neigborhood, admin, country)
        #    1.d extract extract extended tweet, if available, and extended entities, if available
        # 2. decide if post is reaction (reply, quote, share, see https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/entities-object.html)
        # 3. if post is reaction, copy reduced reaction attributes from extracted lbsnPost
        # 4. add post/reaction to recordDict
        # 5. process all referenced posts
        #    5.a Retweet(=Share) and Quote Tweets are special kinds of Tweets that contain the original Tweet as an embedded object.
        #    5.b Retweets have a top-level "retweeted_status" object, and Quoted Tweets have a "quoted_status" object
        # process tweet-post object
        postRecord = self.extractPost(jsonStringDict, userPkey)
        if postRecord is None:
            # in case no post has been extracted (e.g. non_geotagged clause)
            return
        # Assignment Step
        # check if post is reaction to other post
        # reaction means: reduced structure compared to post;
        # reactions often include the complete original post, therefore nested processing necessary
        if HelperFunctions.isPostReaction(jsonStringDict):
            if self.map_reactions is False:
                return
            postReactionRecord = self.mapPostRecord_To_PostReactionRecord(postRecord)
            refUser_pkey = None
            if 'quoted_status' in jsonStringDict: #Quote is both: Share & Reply
                if not 'user' in jsonStringDict.get('quoted_status'):
                    refUser_pkey = HelperFunctions.substituteReferencedUser(jsonStringDict,self.origin,self.log)
                postReactionRecord.reaction_type = lbsnPostReaction.QUOTE
                refPostRecord = self.extractPost(jsonStringDict.get('quoted_status'))
            elif 'retweeted_status' in jsonStringDict:
                # Note: No retweets are available when data is queried using Bounding Box because of Geo-Tweet limitation:
                # "Note that native Retweets are not matched by this parameter. While the original Tweet may have a location, the Retweet will not"
                # see https://developer.twitter.com/en/docs/tweets/filter-realtime/guides/basic-stream-parameters.html
                if not 'user' in jsonStringDict.get('retweeted_status'):
                    # Current issue with Twitter search: the retweeting user is not returned in retweeted_status
                    # but we can get this from other information, such as user_mentions field from the retweet
                    # https://twittercommunity.com/t/status-retweeted-status-quoted-status-user-missing-from-search-tweets-json-response/63355
                    refUser_pkey = HelperFunctions.substituteReferencedUser(jsonStringDict,self.origin,self.log)
                postReactionRecord.reaction_type = lbsnPostReaction.SHARE
                retweetPost = jsonStringDict.get('retweeted_status')
                refPostRecord = self.extractPost(retweetPost, refUser_pkey)

            elif jsonStringDict.get('in_reply_to_status_id_str'):
                #if jsonStringDict.get('in_reply_to_status_id_str') == '778121849687465984':
                #    self.log.warning(f'No User found for status: {jsonStringDict}')
                #    input("Press Enter to continue... (no status will be processed)")

                # if reply, original tweet is not available (?)
                postReactionRecord.reaction_type = lbsnPostReaction.COMMENT
                refPostRecord = HelperFunctions.createNewLBSNRecord_with_id(lbsnPost(),jsonStringDict.get('in_reply_to_status_id_str'),self.origin)
                refUserRecord = HelperFunctions.createNewLBSNRecord_with_id(lbsnUser(),jsonStringDict.get('in_reply_to_user_id_str'),self.origin)
                refUserRecord.user_name = jsonStringDict.get('in_reply_to_screen_name') # Needs to be saved
                self.lbsnRecords.AddRecordsToDict(refUserRecord)
                refPostRecord.user_pkey.CopyFrom(refUserRecord.pkey)

            # add referenced post pkey to reaction
            if not self.disableReactionPostReferencing:
                postReactionRecord.referencedPost_pkey.CopyFrom(refPostRecord.pkey)
                # ToDo: if a Reaction refers to another reaction (Information Spread)
                # This information is currently not [available from Twitter](https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/tweet-object):
                # "Note that retweets of retweets do not show representations of the intermediary retweet [...]"
                # would be added to. postReactionRecord.referencedPostReaction_pkey
                if refPostRecord:
                    self.lbsnRecords.AddRecordsToDict(refPostRecord)
            # add postReactionRecord to Dict
            self.lbsnRecords.AddRecordsToDict(postReactionRecord)
        else:
            # add postReactionRecord to Dict
            self.lbsnRecords.AddRecordsToDict(postRecord)

    def extractUser(self, jsonStringDict):
        user = jsonStringDict
        userRecord = HelperFunctions.createNewLBSNRecord_with_id(lbsnUser(),user.get('id_str'),self.origin)
        # get additional information about the user, if available
        userRecord.user_fullname = user.get('name')
        userRecord.follows = user.get('friends_count')
        userRecord.is_private = user.get('protected')
        userRecord.followed = user.get('followers_count')
        userBio = user.get('description')
        if userBio:
            userRecord.biography = userBio
        userRecord.user_name = user.get('screen_name')
        listedCount = user.get('listed_count')
        if listedCount:
            userRecord.group_count = listedCount
        userRecord.post_count = user.get('statuses_count')
        userRecord.url = f'https://twitter.com/intent/user?user_id={userRecord.pkey.id}'
        refUserLanguage = Language()
        refUserLanguage.language_short = user.get('lang')
        userRecord.user_language.CopyFrom(refUserLanguage)
        userLocation = user.get('location')
        if userLocation:
            userRecord.user_location = userLocation
            if self.geocodes and userRecord.user_location in self.geocodes:
                l_lat = self.geocodes[userRecord.user_location][0]
                l_lng = self.geocodes[userRecord.user_location][1]
                userRecord.user_location_geom = "POINT(%s %s)" % (l_lng,l_lat)
        #userGeoLocation = user.get('profile_location') # todo!
        userRecord.liked_count = user.get('favourites_count')
        userRecord.active_since.CopyFrom(HelperFunctions.parseJSONDateStringToProtoBuf(user.get('created_at')))
        userProfileImageURL = user.get('profile_image_url')
        if not userProfileImageURL == "http://abs.twimg.com/sticky/default_profile_images/default_profile_normal.png":
            userRecord.profile_image_url = userProfileImageURL
        userTimezone = user.get('time_zone')
        if userTimezone:
            userRecord.user_timezone = userTimezone
        userUTCOffset = user.get('utc_offset')
        if userUTCOffset:
            userRecord.user_utc_offset = userUTCOffset
        # the following cannot be extracted from twitter post data
        #deutscherBundestagGroup = HelperFunctions.createNewLBSNRecord_with_id(lbsnUserGroup(),"MdB (Bundestag)",self.origin)
        #userRecord.user_groups_member.append(deutscherBundestagGroup.pkey.id)
        #if self.mapFullRelations:
        #        relationshipRecord = HelperFunctions.createNewLBSNRelationship_with_id(lbsnRelationship(),userRecord.pkey.id,deutscherBundestagGroup.pkey.id, self.origin)
        #        relationshipRecord.relationship_type = lbsnRelationship.inGROUP
        #        self.lbsnRecords.AddRelationshipToDict(relationshipRecord)
        #userRecord.user_groups_follows = []
        return userRecord

    def extractPost(self,jsonStringDict, userPkey = None):
        post_guid = jsonStringDict.get('id_str')
        if not HelperFunctions.check_notice_empty_post_guid(post_guid):
           return None
        postRecord = HelperFunctions.createNewLBSNRecord_with_id(lbsnPost(),post_guid,self.origin)
        postGeoaccuracy = None
        userRecord = None
        userInfo = jsonStringDict.get('user')
        if userInfo :
            # Get Post/Reaction Details of User
            userRecord = self.extractUser(jsonStringDict.get('user'))
        elif userPkey:
            # userPkey is already available for posts that are statuses
            userRecord = HelperFunctions.createNewLBSNRecord_with_id(lbsnUser(),userPkey.id,self.origin)
        if userRecord:
            self.lbsnRecords.AddRecordsToDict(userRecord)
        else:
            self.log.warning(f'Record {self.lbsnRecords.CountGlob}: No User record found for post: {post_guid} (post saved without userid)..')
            print(f'Record {self.lbsnRecords.CountGlob}', end='\r')
            #self.log.warning(f'{originalString}')
            #input("Press Enter to continue... (post will be saved without userid)")

        # Some preprocessing for all types:
        post_coordinates = jsonStringDict.get('coordinates')
        if post_coordinates:
            l_lng = post_coordinates.get('coordinates')[0]
            l_lat = post_coordinates.get('coordinates')[1]
            postRecord.post_geoaccuracy = lbsnPost.LATLNG
            postRecord.post_latlng = "POINT(%s %s)" % (l_lng,l_lat)

        # Check if Place is mentioned
        postPlace_json = jsonStringDict.get('place')
        if postPlace_json:
            # we need some information from postRecord to create placeRecord (e.g. user language, geoaccuracy, post_latlng)
            # some of the information from place will also modify postRecord attributes; therefore return both
            if userRecord:
                userLang = userRecord.user_language
            else:
                userLang = None
            placeRecord, postGeoaccuracy, postCountry = self.extractPlace(postPlace_json, postRecord.post_geoaccuracy, userLang)
            if not postRecord.post_geoaccuracy:
                postRecord.post_geoaccuracy = postGeoaccuracy
            #postRecord.post_geoaccuracy = twitterPostAttributes.geoaccuracy
            self.lbsnRecords.AddRecordsToDict(placeRecord)
            if postCountry:
                postRecord.country_pkey.CopyFrom(postCountry.pkey)
            if isinstance(placeRecord, lbsnCity):
                postRecord.city_pkey.CopyFrom(placeRecord.pkey)
            # either city or place, Twitter user cannot attach both (?)
            elif isinstance(placeRecord, lbsnPlace):
                postRecord.place_pkey.CopyFrom(placeRecord.pkey)
            # substitute postRecord LatLng Coordinates from placeRecord, if not already set
            if not postRecord.post_latlng and postGeoaccuracy in (lbsnPost.CITY,lbsnPost.PLACE):
                postRecord.post_latlng = placeRecord.geom_center
        # if still no geoinformation, send post to Null-Island
        if not postRecord.post_latlng:
            if self.ignore_non_geotagged is True:
                return None
            else:
                self.null_island += 1
                postRecord.post_latlng = "POINT(%s %s)" % (0,0)

        # Process attributes of twitter post
        postSource = jsonStringDict.get('source')
        if postSource:
            postRecord.input_source = HelperFunctions.cleanhtml(jsonStringDict.get('source'))
            if postRecord.input_source in self.ignore_sources_set:
                # skip entry if in ignore list
                return None
        postRecord.post_publish_date.CopyFrom(HelperFunctions.parseJSONDateStringToProtoBuf(jsonStringDict.get('created_at')))
        if userRecord:
            postRecord.user_pkey.CopyFrom(userRecord.pkey)
        valueCount = lambda x: 0 if x is None else x
        postRecord.post_quote_count = valueCount(jsonStringDict.get('quote_count'))
        postRecord.post_comment_count = valueCount(jsonStringDict.get('reply_count'))
        postRecord.post_share_count = valueCount(jsonStringDict.get('retweet_count'))
        postRecord.post_like_count = valueCount(jsonStringDict.get('favorite_count'))
        postRecord.post_url = f'https://twitter.com/statuses/{post_guid}'
        postLanguage = Language()
        postLanguage.language_short = jsonStringDict.get('lang')
        postRecord.post_language.CopyFrom(postLanguage)
        # If Extended_tweet object is available, process entities and post_body (text) data from extended object
        isTruncated = jsonStringDict.get('truncated')
        if isTruncated and 'extended_tweet' in jsonStringDict:
            #if the "truncated" field is set to true, and the "extended_tweet" object provides complete "full_text" and "entities" Tweet metadata
            jsonStringDict = jsonStringDict.get('extended_tweet') # Source for all data is extended object, if available
            postRecord.post_body = jsonStringDict.get('full_text')
            #else:
            #    self.log.warning(f'Truncated but no extended_tweet: {jsonStringDict}')
            #    input("Press Enter to continue... (entry will be skipped)")
            #    return None
        else:
            if 'full_text' in jsonStringDict:
                postRecord.post_body = jsonStringDict.get('full_text')
            else:
                postRecord.post_body = jsonStringDict.get('text')
        #if 'RT @' in postRecord.post_body:
        #    sys.exit(jsonStringDict)
        if 'extended_entities' in jsonStringDict:
            entitiesJson = jsonStringDict.get('extended_entities')
        else:
            entitiesJson = jsonStringDict.get('entities')

        hashtags_json = entitiesJson.get('hashtags')
        if hashtags_json:
            for hashtag in hashtags_json:    #iterate over the list
                postRecord.hashtags.append(hashtag["text"])
        # Look for mentioned userRecords
        userMentionsJson = entitiesJson.get('user_mentions')
        if userMentionsJson:
            refUserRecords = HelperFunctions.getMentionedUsers(userMentionsJson,self.origin)
            self.lbsnRecords.AddRecordsToDict(refUserRecords)
            postRecord.user_mentions_pkey.extend([userRef.pkey for userRef in refUserRecords])
            if self.mapFullRelations:
                for mentionedUserRecord in refUserRecords:
                    relationshipRecord = HelperFunctions.createNewLBSNRelationship_with_id(lbsnRelationship(),userRecord.pkey.id,mentionedUserRecord.pkey.id, self.origin)
                    relationshipRecord.relationship_type = lbsnRelationship.MENTIONS_USER
                    self.lbsnRecords.AddRelationshipToDict(relationshipRecord)
        mediaJson = entitiesJson.get('media')
        if mediaJson:
            postRecord.post_type = HelperFunctions.assignMediaPostType(mediaJson)
        else:
            postRecord.post_type = lbsnPost.TEXT
        postRecord.emoji.extend(HelperFunctions.extract_emoji(postRecord.post_body))
        # because standard print statement will produce escaped text, we can use protobuf text_format to give us a human friendly version of the text
        # log.debug(f'Post record: {text_format.MessageToString(postRecord,as_utf8=True)}')
        # log.debug(f'Post record: {postRecord}')
        return postRecord

    def mapPostRecord_To_PostReactionRecord(self,postRecord):
        postReactionRecord = lbsnPostReaction()
        postReactionRecord.pkey.CopyFrom(postRecord.pkey)
        postReactionRecord.user_pkey.CopyFrom(postRecord.user_pkey)
        postReactionRecord.reaction_latlng = postRecord.post_latlng
        postReactionRecord.reaction_date.CopyFrom(postRecord.post_publish_date) #better post_create_date, but not available from Twitter
        postReactionRecord.reaction_like_count = postRecord.post_like_count
        postReactionRecord.reaction_content = postRecord.post_body
        postReactionRecord.user_mentions_pkey.extend([userRefPkey for userRefPkey in postRecord.user_mentions_pkey])
        return postReactionRecord

    def extractPlace(self, postPlace_json, postGeoaccuracy, userLanguage = None):
        place = postPlace_json
        placeID = place.get('id')
        if not placeID:
           log.warning(f'No PlaceGuid\n\n{postPlace_json}')
           input("Press Enter to continue... (entry will be skipped)")
           return None,postGeoaccuracy,None
        bounding_box_points = place.get('bounding_box').get('coordinates')[0]
        limYMin,limYMax,limXMin,limXMax = HelperFunctions.getRectangleBounds(bounding_box_points)
        bound_points_shapely = geometry.MultiPoint([(limXMin, limYMin), (limXMax, limYMax)])
        lon_center = bound_points_shapely.centroid.coords[0][0] #True centroid (coords may be multipoint)
        lat_center = bound_points_shapely.centroid.coords[0][1]
        place_type = place.get('place_type')
        if place_type == "country":
            # country_guid
            # in case of country, we do not need to save the GUID from Twitter - country_code is already unique
            countryCode = place.get('country_code')
            if countryCode:
                placeRecord = HelperFunctions.createNewLBSNRecord_with_id(lbsnCountry(),place.get('country_code'),self.origin)
                if not postGeoaccuracy:
                    postGeoaccuracy = lbsnPost.COUNTRY
            #else:
            #    log.warning(f'No country_code\n\n{postPlace_json}')
            #    input("Press Enter to continue... (entry will be skipped)")
            #    return None,postGeoaccuracy,None
        elif place_type in ("city","neighborhood","admin"):
            # city_guid
            placeRecord = HelperFunctions.createNewLBSNRecord_with_id(lbsnCity(),place.get('id'),self.origin)
            if not place_type == "city":
                placeRecord.sub_type = place_type
            if not postGeoaccuracy or postGeoaccuracy == lbsnPost.COUNTRY:
                postGeoaccuracy = lbsnPost.CITY
        elif place_type == "poi":
            # place_guid
            # For POIs, City is not available on Twitter
            placeRecord = HelperFunctions.createNewLBSNRecord_with_id(lbsnPlace(),place.get('id'),self.origin)
            if not postGeoaccuracy or postGeoaccuracy in (lbsnPost.COUNTRY, lbsnPost.CITY):
                postGeoaccuracy = lbsnPost.PLACE
        else:
            log.warning(f'No Place Type Detected: {jsonStringDict}')
        #for some reason, twitter place entities sometimes contain linebreaks or whitespaces. We don't want this.
        placeName = place.get('name').replace('\n\r','')
        placeName = re.sub(' +',' ',placeName) # remove multiple whitespace
        if place_type == "poi" or (userLanguage is None or not userLanguage.language_short or userLanguage.language_short in ('en','und')):
            # At the moment, English name are the main references; all other language specific references are stored in name_alternatives - except for places, where twitter has no alternative place names
            # Bugfix necessary: some English names get still saved as name_alternatives
            placeRecord.name = placeName
        else:
            placeRecord.name_alternatives.append(placeName)
        placeRecord.url = place.get('url')
        placeRecord.geom_center = "POINT(%s %s)" % (lon_center,lat_center)
        placeRecord.geom_area = Polygon(bounding_box_points).wkt # prints: 'POLYGON ((0 0, 1 0, 1 1, 0 1, 0 0))'
        refCountryRecord = None
        if not isinstance(placeRecord, lbsnCountry):
            refCountryCode = place.get('country_code')
            if refCountryCode:
                refCountryRecord = HelperFunctions.createNewLBSNRecord_with_id(lbsnCountry(),refCountryCode,self.origin)
                # At the moment, only English name references are processed
                if userLanguage is None or not userLanguage.language_short or userLanguage.language_short in ('en','und'):
                    refCountryRecord.name = place.get('country') # Needs to be saved
                else:
                    alt_name = place.get('country')
                    refCountryRecord.name_alternatives.append(alt_name)
                self.lbsnRecords.AddRecordsToDict(refCountryRecord)
        if postGeoaccuracy == lbsnPost.CITY and refCountryRecord:
            # country_pkey only on lbsnCity(), lbsnPlace() has city_pkey, but this is not available for Twitter
            placeRecord.country_pkey.CopyFrom(refCountryRecord.pkey)
        # log.debug(f'Final Place Record: {placeRecord}')
        return placeRecord, postGeoaccuracy, refCountryRecord


# -*- coding: utf-8 -*-

import psycopg2
from psycopg2 import sql
from .helper_functions import HelperFunctions
from .helper_functions import LBSNRecordDicts
from .store_csv import LBSNcsv
from .shared_structure_proto_lbsndb import ProtoLBSM_db_Mapping
from lbsnstructure.lbsnstructure_pb2 import *
from google.protobuf.timestamp_pb2 import Timestamp
import logging
from sys import exit
import traceback
import os
import sys
# for debugging only:
from google.protobuf import text_format

from glob import glob
from google.protobuf.internal.containers import RepeatedCompositeFieldContainer

class LBSNTransfer():
    def __init__(self, dbCursor = None,
                 dbConnection = None,
                 commit_volume = 10000,
                 disableReactionPostReferencing = 0, storeCSV = None, SUPPRESS_LINEBREAKS = True):
        self.dbCursor = dbCursor
        self.dbConnection = dbConnection
        if not self.dbCursor:
            print("CSV Output Mode.")
        self.count_entries_commit = 0
        self.count_entries_store = 0
        self.count_affected = 0
        self.commit_volume = commit_volume
        self.store_volume = 500000
        self.count_glob = 0
        self.null_island_count = 0
        self.disableReactionPostReferencing = disableReactionPostReferencing
        self.log = logging.getLogger('__main__')
        self.batchedRecords = {lbsnCountry.DESCRIPTOR.name: list(),
                         lbsnCity.DESCRIPTOR.name: list(),
                         lbsnPlace.DESCRIPTOR.name: list(),
                         lbsnUser.DESCRIPTOR.name: list(),
                         lbsnUserGroup.DESCRIPTOR.name: list(),
                         lbsnPost.DESCRIPTOR.name: list(),
                         lbsnPostReaction.DESCRIPTOR.name: list(),
                         lbsnRelationship.DESCRIPTOR.name: list()}

        self.countRound = 0
        self.batchDBVolume = 100 # Records are batched and submitted in one insert with x number of records
        self.storeCSV = storeCSV
        self.headersWritten = set()
        self.db_mapping = ProtoLBSM_db_Mapping()
        #self.CSVsuppressLinebreaks = CSVsuppressLinebreaks

        if self.storeCSV:
            self.csv_output = LBSNcsv(SUPPRESS_LINEBREAKS)

    def commitChanges(self):
        if self.dbCursor:
            self.dbConnection.commit() # commit changes to db
            self.count_entries_commit = 0

    def storeChanges(self):
        if self.storeCSV:
            self.csv_output.clean_csv_batches(self.batchedRecords)
            self.count_entries_store = 0

    def storeLbsnRecordDicts(self, fieldMappingTwitter):
        # order is important here, as PostGres will reject any records where Foreign Keys are violated
        # therefore, records are processed starting from lowest granularity. Order is stored in allDicts()
        self.countRound += 1
        #self.headersWritten.clear()
        recordDicts = fieldMappingTwitter.lbsnRecords
        x = 0
        self.count_affected = 0
        for recordsDict in recordDicts.allDicts:
            type_name = recordsDict[1]
            for record_pkey, record in recordsDict[0].items():
                x += 1
                print(f'Storing {x} of {recordDicts.CountGlob} output records ({type_name})..', end='\r')
                self.prepareLbsnRecord(record,type_name)
                self.count_glob +=  1 #self.dbCursor.rowcount
                self.count_entries_commit +=  1 #self.dbCursor.rowcount
                self.count_entries_store += 1
                if self.dbCursor and (self.count_glob == 100 or self.count_entries_commit > self.commit_volume):
                    self.commitChanges()
                if self.storeCSV and (self.count_entries_store > self.store_volume):
                    self.storeChanges()
        # submit remaining rest
        self.submitAllBatches()
        #self.count_affected += x # monitoring
        print(f'\nRound {self.countRound:03d}: Updated/Inserted {self.count_glob} records.')

    def prepareLbsnRecord(self, record, record_type):
        # clean duplicates in repeated Fields and Sort List
        self.sortCleanProtoRepeatedField(record)
        # store cleaned ProtoBuf records
        self.batchedRecords[record_type].append(record)
        for listType, batchList in self.batchedRecords.items():
            # if any dict contains more values than self.batchDBVolume, submit/store all
            if len(batchList) >= self.batchDBVolume:
                self.submitAllBatches()



    def submitAllBatches(self):
        for recordType, batchList in self.batchedRecords.items():
            if batchList:
                #if self.storeCSV and not recordType in self.headersWritten:
                #    self.writeCSVHeader(recordType)
                #    self.headersWritten.add(recordType)
                self.submitLbsnRecords(recordType)
                #self.funcSubmitSelector(recordType)
                batchList.clear()

    def prepareRecords(self, recordType):
        preparedRecords = []
        for record in self.batchedRecords[recordType]:
            preparedRecord = self.db_mapping.func_prepare_selector(record)
            if preparedRecord:
                preparedRecords.append(preparedRecord)
        return preparedRecords

    def submitLbsnRecords(self, recordType):
        preparedRecords = self.prepareRecords(recordType)
        if self.storeCSV:
            self.csv_output.store_append_batch_to_csv(self.batchedRecords[recordType], self.countRound, recordType)
        if self.dbCursor:
            values_str = ','.join([self.prepareSQLEscapedValues(record) for record in preparedRecords])
            if recordType == lbsnCountry().DESCRIPTOR.name:
                insert_sql = f'''
                            INSERT INTO data."country" ({self.db_mapping.get_header_for_type(recordType)})
                            VALUES {values_str}
                            ON CONFLICT (origin_id,country_guid)
                            DO UPDATE SET
                                name = COALESCE(EXCLUDED.name, data."country".name),
                                name_alternatives = COALESCE((SELECT array_remove(altNamesNewArray,data."country".name) from extensions.mergeArrays(EXCLUDED.name_alternatives, data."country".name_alternatives) AS altNamesNewArray), ARRAY[]::text[]),
                                geom_center = COALESCE(EXCLUDED.geom_center, data."country".geom_center),
                                geom_area = COALESCE(EXCLUDED.geom_area, data."country".geom_area),
                                url = COALESCE(EXCLUDED.url, data."country".url);
                            '''
                            # Array merge of alternatives:
                            # Arrays cannot be null, therefore COALESCE([if array not null],[otherwise create empty array])
                            # We don't want the english name to appear in alternatives, therefore: array_remove(altNamesNewArray,"country".name)
                            # Finally, merge New Entries with existing ones (distinct): extensions.mergeArrays([new],[old]) uses custom mergeArrays function (see function definitions)
            elif recordType == lbsnCity().DESCRIPTOR.name:
                insert_sql = f'''
                            INSERT INTO data."city" ({self.db_mapping.get_header_for_type(recordType)})
                            VALUES {values_str}
                            ON CONFLICT (origin_id,city_guid)
                            DO UPDATE SET
                                name = COALESCE(EXCLUDED.name, data."city".name),
                                name_alternatives = COALESCE((SELECT array_remove(altNamesNewArray,data."city".name) from extensions.mergeArrays(EXCLUDED.name_alternatives, data."city".name_alternatives) AS altNamesNewArray), ARRAY[]::text[]),
                                geom_center = COALESCE(EXCLUDED.geom_center, data."city".geom_center),
                                geom_area = COALESCE(EXCLUDED.geom_area, data."city".geom_area),
                                url = COALESCE(EXCLUDED.url, data."city".url),
                                country_guid = COALESCE(EXCLUDED.country_guid, data."city".country_guid),
                                sub_type = COALESCE(EXCLUDED.sub_type, data."city".sub_type);
                            '''
            elif recordType == lbsnPlace().DESCRIPTOR.name:
                insert_sql = f'''
                            INSERT INTO data."place" ({self.db_mapping.get_header_for_type(recordType)})
                            VALUES {values_str}
                            ON CONFLICT (origin_id,place_guid)
                            DO UPDATE SET
                                name = COALESCE(EXCLUDED.name, data."place".name),
                                name_alternatives = COALESCE((SELECT array_remove(altNamesNewArray,data."place".name) from extensions.mergeArrays(EXCLUDED.name_alternatives, data."place".name_alternatives) AS altNamesNewArray), ARRAY[]::text[]),
                                geom_center = COALESCE(EXCLUDED.geom_center, data."place".geom_center),
                                geom_area = COALESCE(EXCLUDED.geom_area, data."place".geom_area),
                                url = COALESCE(EXCLUDED.url, data."place".url),
                                city_guid = COALESCE(EXCLUDED.city_guid, data."place".city_guid),
                                post_count = GREATEST(COALESCE(EXCLUDED.post_count, data."place".post_count), COALESCE(data."place".post_count, EXCLUDED.post_count));
                            '''
            elif recordType == lbsnUserGroup().DESCRIPTOR.name:
                insert_sql = f'''
                            INSERT INTO data."user_groups" ({self.db_mapping.get_header_for_type(recordType)})
                            VALUES {values_str}
                            ON CONFLICT (origin_id, usergroup_guid)
                            DO UPDATE SET
                                usergroup_name = COALESCE(EXCLUDED.usergroup_name, data."user_groups".usergroup_name),
                                usergroup_description = COALESCE(EXCLUDED.usergroup_description, data."user_groups".usergroup_description),
                                member_count = GREATEST(COALESCE(EXCLUDED.member_count, data."user_groups".member_count), COALESCE(data."user_groups".member_count, EXCLUDED.member_count)),
                                usergroup_createdate = COALESCE(EXCLUDED.usergroup_createdate, data."user_groups".usergroup_createdate),
                                user_owner = COALESCE(EXCLUDED.user_owner, data."user_groups".user_owner);
                            '''
                            # No coalesce for user: in case user changes or removes information, this should also be removed from the record
            elif recordType == lbsnUser().DESCRIPTOR.name:
                insert_sql = f'''
                            INSERT INTO data."user" ({self.db_mapping.get_header_for_type(recordType)})
                            VALUES {values_str}
                            ON CONFLICT (origin_id, user_guid)
                            DO UPDATE SET
                                user_name = COALESCE(EXCLUDED.user_name, data."user".user_name),
                                user_fullname = COALESCE(EXCLUDED.user_fullname, data."user".user_fullname),
                                follows = GREATEST(COALESCE(EXCLUDED.follows, data."user".follows), COALESCE(data."user".follows, EXCLUDED.follows)),
                                followed = GREATEST(COALESCE(EXCLUDED.followed, data."user".followed), COALESCE(data."user".followed, EXCLUDED.followed)),
                                group_count = GREATEST(COALESCE(EXCLUDED.group_count, data."user".group_count), COALESCE(data."user".group_count, EXCLUDED.group_count)),
                                biography = COALESCE(EXCLUDED.biography, data."user".biography),
                                post_count = GREATEST(COALESCE(EXCLUDED.post_count, "user".post_count), COALESCE(data."user".post_count, EXCLUDED.post_count)),
                                is_private = COALESCE(EXCLUDED.is_private, data."user".is_private),
                                url = COALESCE(EXCLUDED.url, data."user".url),
                                is_available = COALESCE(EXCLUDED.is_available, data."user".is_available),
                                user_language = COALESCE(EXCLUDED.user_language, data."user".user_language),
                                user_location = COALESCE(EXCLUDED.user_location, data."user".user_location),
                                user_location_geom = COALESCE(EXCLUDED.user_location_geom, data."user".user_location_geom),
                                liked_count = GREATEST(COALESCE(EXCLUDED.liked_count, data."user".liked_count), COALESCE(data."user".liked_count, EXCLUDED.liked_count)),
                                active_since = COALESCE(EXCLUDED.active_since, data."user".active_since),
                                profile_image_url = COALESCE(EXCLUDED.profile_image_url, data."user".profile_image_url),
                                user_timezone = COALESCE(EXCLUDED.user_timezone, data."user".user_timezone),
                                user_utc_offset = COALESCE(EXCLUDED.user_utc_offset, data."user".user_utc_offset),
                                user_groups_member = COALESCE(extensions.mergeArrays(EXCLUDED.user_groups_member, data."user".user_groups_member), ARRAY[]::text[]),
                                user_groups_follows = COALESCE(extensions.mergeArrays(EXCLUDED.user_groups_follows, data."user".user_groups_follows), ARRAY[]::text[]);
                            '''
            elif recordType == lbsnPost().DESCRIPTOR.name:
                insert_sql = f'''
                            INSERT INTO data."post" ({self.db_mapping.get_header_for_type(recordType)})
                            VALUES {values_str}
                            ON CONFLICT (origin_id, post_guid)
                            DO UPDATE SET
                                post_latlng = COALESCE(EXCLUDED.post_latlng, data."post".post_latlng),
                                place_guid = COALESCE(EXCLUDED.place_guid, data."post".place_guid),
                                city_guid = COALESCE(EXCLUDED.city_guid, data."post".city_guid),
                                country_guid = COALESCE(EXCLUDED.country_guid, data."post".country_guid),
                                post_geoaccuracy = COALESCE(EXCLUDED.post_geoaccuracy, data."post".post_geoaccuracy),
                                user_guid = COALESCE(EXCLUDED.user_guid, data."post".user_guid),
                                post_create_date = COALESCE(EXCLUDED.post_create_date, data."post".post_create_date),
                                post_publish_date = COALESCE(EXCLUDED.post_publish_date, data."post".post_publish_date),
                                post_body = COALESCE(EXCLUDED.post_body, data."post".post_body),
                                post_language = COALESCE(EXCLUDED.post_language, data."post".post_language),
                                user_mentions = COALESCE(EXCLUDED.user_mentions, data."post".user_mentions),
                                hashtags = COALESCE(extensions.mergeArrays(EXCLUDED.hashtags, data."post".hashtags), ARRAY[]::text[]),
                                emoji = COALESCE(extensions.mergeArrays(EXCLUDED.emoji, data."post".emoji), ARRAY[]::text[]),
                                post_like_count = COALESCE(EXCLUDED.post_like_count, data."post".post_like_count),
                                post_comment_count = COALESCE(EXCLUDED.post_comment_count, data."post".post_comment_count),
                                post_views_count = COALESCE(EXCLUDED.post_views_count, data."post".post_views_count),
                                post_title = COALESCE(EXCLUDED.post_title, data."post".post_title),
                                post_thumbnail_url = COALESCE(EXCLUDED.post_thumbnail_url, data."post".post_thumbnail_url),
                                post_url = COALESCE(EXCLUDED.post_url, data."post".post_url),
                                post_type = COALESCE(EXCLUDED.post_type, data."post".post_type),
                                post_filter = COALESCE(EXCLUDED.post_filter, data."post".post_filter),
                                post_quote_count = COALESCE(EXCLUDED.post_quote_count, data."post".post_quote_count),
                                post_share_count = COALESCE(EXCLUDED.post_share_count, data."post".post_share_count),
                                input_source = COALESCE(EXCLUDED.input_source, data."post".input_source),
                                post_content_license = COALESCE(EXCLUDED.post_content_license, data."post".post_content_license);
                            '''
            elif recordType == lbsnPostReaction().DESCRIPTOR.name:
                insert_sql = f'''
                            INSERT INTO data."post_reaction" ({self.db_mapping.get_header_for_type(recordType)})
                            VALUES {values_str}
                            ON CONFLICT (origin_id, reaction_guid)
                            DO UPDATE SET
                                reaction_latlng = COALESCE(EXCLUDED.reaction_latlng, data."post_reaction".reaction_latlng),
                                user_guid = COALESCE(EXCLUDED.user_guid, data."post_reaction".user_guid),
                                referencedPost_guid = COALESCE(EXCLUDED.referencedPost_guid, data."post_reaction".referencedPost_guid),
                                referencedPostreaction_guid = COALESCE(EXCLUDED.referencedPostreaction_guid, data."post_reaction".referencedPostreaction_guid),
                                reaction_type = COALESCE(EXCLUDED.reaction_type, data."post_reaction".reaction_type),
                                reaction_date = COALESCE(EXCLUDED.reaction_date, data."post_reaction".reaction_date),
                                reaction_content = COALESCE(EXCLUDED.reaction_content, data."post_reaction".reaction_content),
                                reaction_like_count = COALESCE(EXCLUDED.reaction_like_count, data."post_reaction".reaction_like_count),
                                user_mentions = COALESCE(extensions.mergeArrays(EXCLUDED.user_mentions, data."post_reaction".user_mentions), ARRAY[]::text[]);
                            '''
            self.submitBatch(insert_sql)

    def submitLbsnRelationships(self):
        # submit relationships of different types record[1] is the PostgresQL formatted list of values, record[0] is the type of relationship that determines the table selection
        selectFriends = [relationship[1] for relationship in self.batchedRecords[lbsnRelationship().DESCRIPTOR.name] if relationship[0] == "isfriend"]
        if selectFriends:
            if self.storeCSV:
                self.csv_output.store_append_batch_to_csv(selectFriends, self.countRound, '_user_friends_user')
            if self.dbCursor:
                args_isFriend = ','.join(selectFriends)
                insert_sql = f'''
                                INSERT INTO relations."_user_friends_user" ({self.typeNamesHeaderDict["_user_friends_user"]})
                                VALUES {args_isFriend}
                                ON CONFLICT (origin_id, user_guid, friend_guid)
                                DO NOTHING
                            '''
                self.submitBatch(insert_sql)
        selectConnected = [relationship[1] for relationship in self.batchedRecords[lbsnRelationship().DESCRIPTOR.name] if relationship[0] == "isconnected"]
        if selectConnected:
            if self.storeCSV:
                self.csv_output.store_append_batch_to_csv(selectConnected, self.countRound, '_user_connectsto_user')
            if self.dbCursor:
                args_isConnected = ','.join(selectConnected)
                insert_sql = f'''
                                INSERT INTO relations."_user_connectsto_user" ({self.typeNamesHeaderDict["_user_connectsto_user"]})
                                VALUES {args_isConnected}
                                ON CONFLICT (origin_id, user_guid, connectedto_user_guid)
                                DO NOTHING
                            '''
                self.submitBatch(insert_sql)
        selectUserGroupMember = [relationship[1] for relationship in self.batchedRecords[lbsnRelationship().DESCRIPTOR.name] if relationship[0] == "ingroup"]
        if selectUserGroupMember:
            if self.storeCSV:
                self.csv_output.store_append_batch_to_csv(selectUserGroupMember, self.countRound, '_user_memberof_group')
            if self.dbCursor:
                args_isInGroup = ','.join(selectUserGroupMember)
                insert_sql = f'''
                                INSERT INTO relations."_user_memberof_group" ({self.typeNamesHeaderDict["_user_memberof_group"]})
                                VALUES {args_isInGroup}
                                ON CONFLICT (origin_id, user_guid, group_guid)
                                DO NOTHING
                            '''
                self.submitBatch(insert_sql)
        selectUserGroupMember = [relationship[1] for relationship in self.batchedRecords[lbsnRelationship().DESCRIPTOR.name] if relationship[0] == "followsgroup"]
        if selectUserGroupMember:
            if self.storeCSV:
               self.csv_output.store_append_batch_to_csv(selectUserGroupMember, self.countRound, '_user_follows_group')
            if self.dbCursor:
                args_isInGroup = ','.join(selectUserGroupMember)
                insert_sql = f'''
                                INSERT INTO relations."_user_follows_group" ({self.typeNamesHeaderDict["_user_follows_group"]})
                                VALUES {args_isInGroup}
                                ON CONFLICT (origin_id, user_guid, group_guid)
                                DO NOTHING
                            '''
                self.submitBatch(insert_sql)
        selectUserMentions = [relationship[1] for relationship in self.batchedRecords[lbsnRelationship().DESCRIPTOR.name] if relationship[0] == "mentions_user"]
        if selectUserMentions:
            if self.storeCSV:
                self.csv_output.store_append_batch_to_csv(selectUserMentions, self.countRound, '_user_mentions_user')
            if self.dbCursor:
                args_isInGroup = ','.join(selectUserMentions)
                insert_sql = f'''
                                INSERT INTO relations."_user_mentions_user" ({self.typeNamesHeaderDict["_user_mentions_user"]})
                                VALUES {args_isInGroup}
                                ON CONFLICT (origin_id, user_guid, mentioneduser_guid)
                                DO NOTHING
                            '''
                self.submitBatch(insert_sql)

    def submitBatch(self,insert_sql):
        ## Needs testing: is using Savepoint for each insert slower than rolling back entire commit?
        ## for performance, see https://stackoverflow.com/questions/12206600/how-to-speed-up-insertion-performance-in-postgresql
        ## or this: https://stackoverflow.com/questions/8134602/psycopg2-insert-multiple-rows-with-one-query
        self.dbCursor.execute("SAVEPOINT submit_recordBatch")
        tsuccessful = False
        while not tsuccessful:
            try:
                self.dbCursor.execute(insert_sql)
            except psycopg2.IntegrityError as e:
                if '(post_language)' in e.diag.message_detail or '(user_language)' in e.diag.message_detail:
                    # If language does not exist, we'll trust Twitter and add this to our language list
                    missingLanguage = e.diag.message_detail.partition("language)=(")[2].partition(") is not present")[0]
                    print(f'TransactionIntegrityError, inserting language "{missingLanguage}" first..               ')
                    self.dbCursor.execute("ROLLBACK TO SAVEPOINT submit_recordBatch")
                    insert_language_sql = '''
                           INSERT INTO data."language" (language_short,language_name,language_name_de)
                           VALUES (%s,NULL,NULL);
                           '''
                    self.dbCursor.execute(insert_language_sql,(missingLanguage,))
                else:
                    sys.exit(f'{e}')
            except ValueError as e:
                self.log.warning(f'{e}')
                input("Press Enter to continue... (entry will be skipped)")
                self.log.warning(f'{insert_sql}')
                input("args:... ")
                self.log.warning(f'{args_str}')
                self.dbCursor.execute("ROLLBACK TO SAVEPOINT submit_recordBatch")
                tsuccessful = True
            except psycopg2.ProgrammingError as e:
                sys.exit(insert_sql)
            else:
                #self.count_affected += self.dbCursor.rowcount # monitoring
                self.dbCursor.execute("RELEASE SAVEPOINT submit_recordBatch")
                tsuccessful = True

    def prepareSQLEscapedValues(self, *args):
        # dynamically construct sql value injection
        # e.g. record_sql = '''(%s,%s,%s,%s,%s,%s,%s)'''
        record_sql = f'''{','.join('%s' for x in range(0, len(args)))}'''
        # inject values
        preparedSQLRecord = self.dbCursor.mogrify(record_sql, tuple(args))
        #mogrify returns a byte object, we decode it so it can be used as a string again
        preparedSQLRecord = preparedSQLRecord.decode()
        return preparedSQLRecord

    def sortCleanProtoRepeatedField(self, record):
        # remove duplicate values in repeated field, sort alphabetically
        # needed for unique compare
        for descriptor in record.DESCRIPTOR.fields:
            if descriptor.label == descriptor.LABEL_REPEATED:
                x = getattr(record, descriptor.name)
                if x and not isinstance(x, RepeatedCompositeFieldContainer):
                    xCleaned = set(x)
                    xSorted = sorted(xCleaned)
                    #Complete clear of repeated field
                    for key in range(0, len(x)):
                        x.pop()
                    # add sorted list
                    x.extend(xSorted)

    def finalize(self):
        """ Final procedure calls:
            - clean and merge csv batches
        """
        if self.storeCSV:
            self.csv_output.clean_csv_batches(self.batchedRecords)




# -*- coding: utf-8 -*-
"""
Tests for LBSNtransfer tool
"""

@staticmethod
def add_bundestag_group_example(twitter_records):
    """ Example: Manually add Group that applies to all entries.
        To use, must uncomment fieldMapping/extractUser
    """

    deutscher_bundestag_group = HF.createNewLBSNRecord_with_id(lbsnUserGroup(), "MdB (Bundestag)", twitter_records.origin)
    dbg_owner = HF.createNewLBSNRecord_with_id(lbsnUser(), "243586130", twitter_records.origin)
    dbg_owner.user_name = 'wahl_beobachter'
    dbg_owner.user_fullname = 'Martin Fuchs'
    deutscher_bundestag_group.user_owner_pkey.CopyFrom(dbg_owner.pkey)
    deutscher_bundestag_group.usergroup_description = 'Alle twitternden Abgeordneten aus dem Deutschen Bundestag #bundestag'
    twitter_records.lbsnRecords.AddRecordToDict(dbg_owner)
    twitter_records.lbsnRecords.AddRecordToDict(deutscher_bundestag_group)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Collection of classes to import, convert and export data
from and to common lbsn structure

For more info, see [concept](https://gitlab.vgiscience.de/lbsn/concept)
"""

from .classes.db_connection import DBConnection
from .classes.helper_functions import HelperFunctions
from .classes.helper_functions import LBSNRecordDicts
from .classes.helper_functions import GeocodeLocations
from .classes.helper_functions import TimeMonitor
from .classes.field_mapping_twitter import FieldMappingTwitter
from .classes.field_mapping_flickr import FieldMappingFlickr
from .classes.shared_structure_proto_lbsndb import ProtoLBSM_db_Mapping
from .classes.submit_data import LBSNTransfer
from .classes.load_data import LoadData
from .classes.store_csv import LBSNcsv
from .config.config import BaseConfig



# -*- coding: utf-8 -*-

from lbsnstructure.lbsnstructure_pb2 import *
from .helper_functions import HelperFunctions as HF

class ProtoLBSM_db_Mapping():

    @staticmethod
    def get_header_for_type(desc_name):
        """ Will return lbsn-db headers for proto buf record types
        """
        dict_switcher = {lbsnCity.DESCRIPTOR.name: 'origin_id, city_guid, name, name_alternatives, geom_center, geom_area, url, country_guid, sub_type',
                                    lbsnCountry.DESCRIPTOR.name: 'origin_id, country_guid, name, name_alternatives, geom_center, geom_area, url',
                                    lbsnPlace.DESCRIPTOR.name: 'origin_id, place_guid, name, name_alternatives, geom_center, geom_area, url, city_guid, post_count',
                                    lbsnPost.DESCRIPTOR.name: 'origin_id, post_guid, post_latlng, place_guid, city_guid, country_guid, post_geoaccuracy, user_guid, post_create_date, post_publish_date, post_body, post_language, user_mentions, hashtags, emoji, post_like_count, post_comment_count, post_views_count, post_title, post_thumbnail_url, post_url, post_type, post_filter, post_quote_count, post_share_count, input_source, post_content_license',
                                    lbsnPostReaction.DESCRIPTOR.name: 'origin_id, reaction_guid, reaction_latlng, user_guid, referencedPost_guid, referencedPostreaction_guid, reaction_type, reaction_date, reaction_content, reaction_like_count, user_mentions',
                                    lbsnUser.DESCRIPTOR.name: 'origin_id, user_guid, user_name, user_fullname, follows, followed, group_count, biography, post_count, is_private, url, is_available, user_language, user_location, user_location_geom, liked_count, active_since, profile_image_url, user_timezone, user_utc_offset, user_groups_member, user_groups_follows',
                                    lbsnUserGroup.DESCRIPTOR.name: 'origin_id, usergroup_guid, usergroup_name, usergroup_description, member_count, usergroup_createdate, user_owner',
                                    '_user_mentions_user': 'origin_id, user_guid, mentioneduser_guid',
                                    '_user_follows_group': 'origin_id, user_guid, group_guid',
                                    '_user_memberof_group': 'origin_id, user_guid, group_guid',
                                    '_user_connectsto_user': 'origin_id, user_guid, connectedto_user_guid',
                                    '_user_friends_user': 'origin_id, user_guid, friend_guid'
                               }
        return dict_switcher.get(desc_name)

    def func_prepare_selector(self, record):
        dictSwitcher = {
            lbsnCountry().DESCRIPTOR.name: self.prepareLbsnCountry,
            lbsnCity().DESCRIPTOR.name: self.prepareLbsnCity,
            lbsnPlace().DESCRIPTOR.name: self.prepareLbsnPlace,
            lbsnUser().DESCRIPTOR.name: self.prepareLbsnUser,
            lbsnUserGroup().DESCRIPTOR.name: self.prepareLbsnUserGroup,
            lbsnPost().DESCRIPTOR.name: self.prepareLbsnPost,
            lbsnPostReaction().DESCRIPTOR.name: self.prepareLbsnPostReaction,
            lbsnRelationship().DESCRIPTOR.name: self.prepareLbsnRelationship
        }
        prepareFunction = dictSwitcher.get(record.DESCRIPTOR.name)
        return prepareFunction(record)

    def prepareLbsnCountry(self, record):
        # Get common attributes for place types Place, City and Country
        placeRecord = placeAttrShared(record)
        preparedRecord = (placeRecord.OriginID,
                          placeRecord.Guid,
                          placeRecord.name,
                          placeRecord.name_alternatives,
                          HF.returnEWKBFromGeoTEXT(placeRecord.geom_center),
                          HF.returnEWKBFromGeoTEXT(placeRecord.geom_area),
                          placeRecord.url)
        return preparedRecord

    def prepareLbsnCity(self, record):
        placeRecord = placeAttrShared(record)
        countryGuid = HF.null_check(record.country_pkey.id)
        subType = HF.null_check(record.sub_type)
        preparedRecord = (placeRecord.OriginID,
                          placeRecord.Guid,
                          placeRecord.name,
                          placeRecord.name_alternatives,
                          HF.returnEWKBFromGeoTEXT(placeRecord.geom_center),
                          HF.returnEWKBFromGeoTEXT(placeRecord.geom_area),
                          placeRecord.url,
                          countryGuid,
                          subType)
        return preparedRecord

    def prepareLbsnPlace(self, record):
        placeRecord = placeAttrShared(record)
        cityGuid = HF.null_check(record.city_pkey.id)
        postCount = HF.null_check(record.post_count)
        preparedRecord = (placeRecord.OriginID,
                          placeRecord.Guid,
                          placeRecord.name,
                          placeRecord.name_alternatives,
                          HF.returnEWKBFromGeoTEXT(placeRecord.geom_center),
                          HF.returnEWKBFromGeoTEXT(placeRecord.geom_area),
                          placeRecord.url,
                          cityGuid,
                          postCount)
        return preparedRecord

    def prepareLbsnUser(self, record):
        userRecord = userAttrShared(record)
        preparedRecord = (userRecord.OriginID,
                          userRecord.Guid,
                          userRecord.user_name,
                          userRecord.user_fullname,
                          userRecord.follows,
                          userRecord.followed,
                          userRecord.group_count,
                          userRecord.biography,
                          userRecord.post_count,
                          userRecord.is_private,
                          userRecord.url,
                          userRecord.is_available,
                          userRecord.user_language,
                          userRecord.user_location,
                          HF.returnEWKBFromGeoTEXT(userRecord.user_location_geom),
                          userRecord.liked_count,
                          userRecord.active_since,
                          userRecord.profile_image_url,
                          userRecord.user_timezone,
                          userRecord.user_utc_offset,
                          userRecord.user_groups_member,
                          userRecord.user_groups_follows)
        return preparedRecord

    def prepareLbsnUserGroup(self, record):
        userGroupRecord = userGroupAttrShared(record)
        preparedRecord = (userGroupRecord.OriginID,
                          userGroupRecord.Guid,
                          userGroupRecord.usergroup_name,
                          userGroupRecord.usergroup_description,
                          userGroupRecord.member_count,
                          userGroupRecord.usergroup_createdate,
                          userGroupRecord.user_owner)
        return preparedRecord

    def prepareLbsnPost(self, record):
        postRecord = postAttrShared(record)
        preparedRecord = (postRecord.OriginID,
                          postRecord.Guid,
                          HF.returnEWKBFromGeoTEXT(postRecord.post_latlng),
                          postRecord.place_guid,
                          postRecord.city_guid,
                          postRecord.country_guid,
                          postRecord.post_geoaccuracy,
                          postRecord.user_guid,
                          postRecord.post_create_date,
                          postRecord.post_publish_date,
                          postRecord.post_body,
                          postRecord.post_language,
                          postRecord.user_mentions,
                          postRecord.hashtags,
                          postRecord.emoji,
                          postRecord.post_like_count,
                          postRecord.post_comment_count,
                          postRecord.post_views_count,
                          postRecord.post_title,
                          postRecord.post_thumbnail_url,
                          postRecord.post_url,
                          postRecord.post_type,
                          postRecord.post_filter,
                          postRecord.post_quote_count,
                          postRecord.post_share_count,
                          postRecord.input_source,
                          postRecord.post_content_license)
        return preparedRecord

    def prepareLbsnPostReaction(self, record):
        postReactionRecord = postReactionAttrShared(record)
        preparedRecord = (postReactionRecord.OriginID,
                          postReactionRecord.Guid,
                          HF.returnEWKBFromGeoTEXT(postReactionRecord.reaction_latlng),
                          postReactionRecord.user_guid,
                          postReactionRecord.referencedPost,
                          postReactionRecord.referencedPostreaction,
                          postReactionRecord.reaction_type,
                          postReactionRecord.reaction_date,
                          postReactionRecord.reaction_content,
                          postReactionRecord.reaction_like_count,
                          postReactionRecord.user_mentions)
        return preparedRecord

    def prepareLbsnRelationship(self, record):
        relationshipRecord = relationshipAttrShared(record)
        preparedTypeRecordTuple = (relationshipRecord.relType,
                          self.prepareRecordValues(relationshipRecord.OriginID,
                          relationshipRecord.Guid,
                          relationshipRecord.Guid_Rel))
        return preparedTypeRecordTuple

class placeAttrShared():
    def __init__(self, record):
        self.OriginID = record.pkey.origin.origin_id # = 3
        self.Guid = record.pkey.id
        self.name = HF.null_check(record.name)
        # because ProtoBuf Repeated Field does not support distinct rule, we remove any duplicates in list fields prior to submission here
        self.name_alternatives = list(set(record.name_alternatives))
        if self.name and self.name in self.name_alternatives:
            self.name_alternatives.remove(self.name)
        self.url = HF.null_check(record.url)
        self.geom_center = HF.null_check(record.geom_center)
        self.geom_area = HF.null_check(record.geom_area)

class userAttrShared():
    def __init__(self, record):
        self.OriginID = record.pkey.origin.origin_id
        self.Guid = record.pkey.id
        self.user_name = HF.null_check(record.user_name)
        self.user_fullname = HF.null_check(record.user_fullname)
        self.follows = HF.null_check(record.follows)
        self.followed = HF.null_check(record.followed)
        self.group_count = HF.null_check(record.group_count)
        self.biography = HF.null_check(record.biography)
        self.post_count = HF.null_check(record.post_count)
        self.url = HF.null_check(record.url)
        self.is_private = HF.null_check(record.is_private)
        self.is_available = HF.null_check(record.is_available)
        self.user_language = HF.null_check(record.user_language.language_short)
        self.user_location = HF.null_check(record.user_location)
        self.user_location_geom = HF.null_check(record.user_location_geom)
        self.liked_count = HF.null_check(record.liked_count)
        self.active_since = HF.null_check_datetime(record.active_since)
        self.profile_image_url = HF.null_check(record.profile_image_url)
        self.user_timezone = HF.null_check(record.user_timezone)
        self.user_utc_offset = HF.null_check(record.user_utc_offset)
        self.user_groups_member = list(set(record.user_groups_member))
        self.user_groups_follows = list(set(record.user_groups_follows))

class userGroupAttrShared():
    def __init__(self, record):
        self.OriginID = record.pkey.origin.origin_id
        self.Guid = record.pkey.id
        self.usergroup_name = HF.null_check(record.usergroup_name)
        self.usergroup_description = HF.null_check(record.usergroup_description)
        self.member_count = HF.null_check(record.member_count)
        self.usergroup_createdate = HF.null_check_datetime(record.usergroup_createdate)
        self.user_owner = HF.null_check(record.user_owner_pkey.id)

class postAttrShared():
    def __init__(self, record):
        self.OriginID = record.pkey.origin.origin_id
        self.Guid = record.pkey.id
        self.post_latlng = HF.null_check(record.post_latlng)
        self.place_guid = HF.null_check(record.place_pkey.id)
        self.city_guid = HF.null_check(record.city_pkey.id)
        self.country_guid = HF.null_check(record.country_pkey.id)
        self.post_geoaccuracy = HF.null_check(lbsnPost().PostGeoaccuracy.Name(record.post_geoaccuracy)).lower()
        self.user_guid = HF.null_check(record.user_pkey.id)
        self.post_create_date = HF.null_check_datetime(record.post_create_date)
        self.post_publish_date = HF.null_check_datetime(record.post_publish_date)
        self.post_body = HF.null_check(record.post_body)
        self.post_language = HF.null_check(record.post_language.language_short)
        self.user_mentions = list(set([pkey.id for pkey in record.user_mentions_pkey]))
        self.hashtags = list(set(record.hashtags))
        self.emoji = list(set(record.emoji))
        self.post_like_count = HF.null_check(record.post_like_count)
        self.post_comment_count = HF.null_check(record.post_comment_count)
        self.post_views_count = HF.null_check(record.post_views_count)
        self.post_title = HF.null_check(record.post_title)
        self.post_thumbnail_url = HF.null_check(record.post_thumbnail_url)
        self.post_url = HF.null_check(record.post_url)
        self.post_type = HF.null_check(lbsnPost().PostType.Name(record.post_type)).lower()
        self.post_filter = HF.null_check(record.post_filter)
        self.post_quote_count = HF.null_check(record.post_quote_count)
        self.post_share_count = HF.null_check(record.post_share_count)
        self.input_source = HF.null_check(record.input_source)
        self.post_content_license = HF.null_check(record.post_content_license)

class postReactionAttrShared():
    def __init__(self, record):
        self.OriginID = record.pkey.origin.origin_id
        self.Guid = record.pkey.id
        self.reaction_latlng = HF.null_check(record.reaction_latlng)
        self.user_guid = HF.null_check(record.user_pkey.id)
        self.referencedPost = HF.null_check(record.referencedPost_pkey.id)
        self.referencedPostreaction = HF.null_check(record.referencedPostreaction_pkey.id)
        self.reaction_type = HF.null_check(lbsnPostReaction().ReactionType.Name(record.reaction_type)).lower()
        self.reaction_date = HF.null_check_datetime(record.reaction_date)
        self.reaction_content = HF.null_check(record.reaction_content)
        self.reaction_like_count = HF.null_check(record.reaction_like_count)
        self.user_mentions = list(set([pkey.id for pkey in record.user_mentions_pkey]))

class relationshipAttrShared():
    def __init__(self, relationship):
        self.OriginID = relationship.pkey.relation_to.origin.origin_id
        self.Guid = relationship.pkey.relation_to.id
        self.Guid_Rel = relationship.pkey.relation_from.id
        self.relType = HF.null_check(lbsnRelationship().RelationshipType.Name(relationship.relationship_type)).lower()


# -*- coding: utf-8 -*-
from .db_connection import DBConnection
from .helper_functions import HelperFunctions as HF
from glob import glob
from json import loads as json_loads, decoder as json_decoder
import os
import ntpath
import csv
from _csv import QUOTE_MINIMAL


class LoadData():

    def loop_input_records(records, transferlimit, import_mapper, config):
        """Loops input json or csv records, converts to ProtoBuf structure and adds to records_dict

        Returns statistic-counts, modifies (adds results to) import_mapper
        """

        finished = False
        processed_records = 0
        db_row_number = 0
        for record in records:
            processed_records += 1
            if config.is_local_input:
                single_record = record
            else:
                db_row_number = record[0]
                single_record = record[2]
            if LoadData.skip_empty_or_other(single_record):
                continue
            if config.local_file_type == 'json' or not config.is_local_input:
                import_mapper.parseJsonRecord(single_record, config.input_lbsn_type)
            elif config.local_file_type in ('txt','csv'):
                import_mapper.parse_csv_record(single_record)
            else:
                exit(f'Format {config.local_file_type} not supportet.')

            if (transferlimit and processed_records >= transferlimit) or \
               (not config.is_local_input and config.end_with_db_row_number and db_row_number >= config.end_with_db_row_number):
                finished = True
                break
        return processed_records, finished

    @staticmethod
    def skip_empty_or_other(single_record):
        """Detect  Rate Limiting Notice or empty records
           so they can be skipped.
        """
        skip = False
        if not single_record or (isinstance(single_record,dict) and single_record.get('limit')):
            skip = True
        return skip

    def fetch_json_data_from_lbsn(cursor, start_id=0, get_max=None, number_of_records_to_fetch=10000):
        """Fetches records from Postgres DB

        Keyword arguments:
        cursor -- db-cursor
        start_id -- Offset for querying
        get_max -- optional limit for retrieving records
        number_of_records_to_fetch -- how many records should get fetched
        """
        # if transferlimit is below number_of_records_to_fetch, e.g.  10000,
        # retrieve only necessary volume of records
        if get_max:
            number_of_records_to_fetch = min(number_of_records_to_fetch, get_max)
        query_sql = '''
                SELECT in_id,insert_time,data::json FROM public."input"
                WHERE in_id > %s
                ORDER BY in_id ASC LIMIT %s;
                '''
        cursor.execute(query_sql, (start_id, number_of_records_to_fetch))
        records = cursor.fetchall()
        if cursor.rowcount == 0:
            return None
        return records

    def fetch_data_from_file(loc_filelist, continue_number, is_stacked_json, format):
        """Fetches CSV or JSON data (including stacked json) from file"""
        if format == 'json':
           records = LoadData.fetch_json_data_from_file(loc_filelist,
                                                     continue_number,
                                                     config.is_stacked_json)
        elif format == 'txt':
           records = LoadData.fetch_csv_data_from_file(loc_filelist,
                                                     continue_number)
        else:
           exit(f'Format {format} not supported.')
        return records

    @staticmethod
    def fetch_json_data_from_file(loc_filelist, start_file_id=0, is_stacked_json=False):
        """Read json entries from file.

        Typical form is [{json1},{json2}], if is_stacked_json is True:
        will process stacked jsons in the form of {json1}{json2}
        """
        records = []
        loc_file = loc_filelist[start_file_id]
        with open(loc_file, 'r', encoding="utf-8", errors='replace') as file:
            # Stacked JSON is a simple file with many concatenated jsons, e.g.
            # {json1}{json2} etc.
            if is_stacked_json:
                try:
                    for obj in HF.decode_stacked(file.read()):
                        records.append(obj)
                except json_decoder.JSONDecodeError:
                    pass
            else:
                # normal json nesting, e.g.  {{record1},{record2}}
                records = json_loads(file.read())
        if records:
            return records
        return None

    @staticmethod
    def fetch_csv_data_from_file(loc_filelist, start_file_id=0):
        """Read csv entries from file (either *.txt or *.csv).

        The actual CSV formatting is not setable in config yet. There are many specifics, e.g.
        #QUOTE_NONE is used here because media saved from Flickr does not contain any quotes ""
        """
        records = []
        loc_file = loc_filelist[start_file_id]
        HF.log_main_debug(f'\nCurrent file: {ntpath.basename(loc_file)}')
        with open(loc_file, 'r', encoding="utf-8", errors='replace') as file:
            reader = csv.reader(file, delimiter=',', quotechar='"', quoting=csv.QUOTE_NONE)
            next(reader, None)  # skip headerline
            records = list(reader)
        if not records:
            return None
        return records

    def initialize_output_connection(config):
        """Establishes connection to output DB (Postgres), if set in config"""
        if config.dbUser_Output:
            output_connection = DBConnection(config.dbServeradressOutput,
                                             config.dbNameOutput,
                                             config.dbUser_Output,
                                             config.dbPassword_Output)
            conn_output, cursor_output = output_connection.connect()
        else:
            conn_output = None
            cursor_output = None
        return conn_output, cursor_output

    def initialize_input_connection(config):
        """Establishes connection to input DB (Postgres)

        Returns cursor
        """
        input_connection = DBConnection(config.dbServeradressInput,
                                        config.dbNameInput,
                                        config.dbUser_Input,
                                        config.dbPassword_Input,
                                        True # ReadOnly Mode
                                        )
        conn_input, cursor_input = input_connection.connect()
        return cursor_input

    def read_local_files(config):
        """Read Local Files according to config parameters and returns list of file-paths"""
        path = f'{config.InputPath}'
        if config.recursiveLoad:
            excludefolderlist = ["01_DataSetHistory","02_UserData","03_ClippedData","04_MapVis"]
            excludestartswithfile = ["log","settings","GridCoordinates"]
            loc_filelist = LoadData.scan_rec(path,
                                             format=config.local_file_type,
                                             excludefolderlist=excludefolderlist,
                                             excludestartswithfile=excludestartswithfile)
        else:
            loc_filelist = glob(f'{path}*.{config.local_file_type}')

        input_count = (len(loc_filelist))
        if input_count == 0:
            sys.exit("No location files found.")
        else:
            return loc_filelist

    def load_geocodes(geo_config):
        """Loads coordinates-string tuples for geocoding text locations (Optional)"""
        locationsgeocode_dict = geocodeLocations()
        locationsgeocode_dict.load_geocodelist(geo_config)
        geocode_dict = locationsgeocode_dict.geocode_dict
        return geocode_dict

    def load_ignore_sources(list_source):
        """Loads list of source types to be ignored"""
        ignore_source_list = set()
        with open(list_source, newline='', encoding='utf8') as f:
            for ignore_source in f:
                ignore_source_list.add(ignore_source.strip())
        return ignore_source_list

    @staticmethod
    def scan_rec(root, subdirlimit=2, format="csv", excludefolderlist=[], excludestartswithfile=[]):
        """Recursively scan subdir for datafiles"""
        rval = []
        def do_scan(start_dir,output,depth=0):
            for f in os.listdir(start_dir):
                ff = os.path.join(start_dir,f)
                if os.path.isdir(ff):
                    if depth<subdirlimit:
                        efound = False
                        for entry in excludefolderlist: #check for excludefolders
                            if entry in ff:
                                efound = True
                                break
                        if efound == False:
                            do_scan(ff,output,depth+1)
                else:
                    if ff.endswith(format):
                        efound = False
                        for entry in excludestartswithfile:
                            if ntpath.basename(ff).startswith(entry):
                                efound = True
                                break
                        if efound == False:
                            output.append(ff)
        do_scan(root,rval,0)
        return rval


# -*- coding: utf-8 -*-

from setuptools import setup
import sys

with open('README.md') as f:
    long_description = f.read()

with open('VERSION') as version_file:
    version_var = version_file.read().strip()     
     
## setuptools dev
setup(  name = "lbsntransform",
        version = version_var,
        description = "Location based social network (LBSN) data structure format & transfer tool",
        long_description=long_description,
        long_description_content_type='text/markdown',
        author='Alexander Dunkel',
        author_email='alexander.dunkel@tu-dresden.de',
        url='https://gitlab.vgiscience.de/lbsn/lbsntransform',
        license='GNU GPLv3 or any higher',
        packages=['lbsntransform'],
        include_package_data=True,
        install_requires=[
            'lbsnstructure>=0.2.6.211',
            'protobuf',
            'psycopg2',
            'ppygis3',
            'shapely',
            'emoji'
        ],
        entry_points={
        'console_scripts': [
            'lbsntransform = lbsntransform.__main__:main'
        ]
        })